;;; dml-mode.el --- Major mode for editing (Dynamic) ML

;; Copyright (C) 2012       Ian Zerny
;; Copyright (C) 1999,2000,2004  Stefan Monnier
;; Copyright (C) 1994-1997  Matthew J. Morley
;; Copyright (C) 1989       Lars Bo Nielsen

;; Author: Lars Bo Nielsen
;;      Olin Shivers
;;	Fritz Knabe (?)
;;	Steven Gilmore (?)
;;	Matthew Morley <mjm@scs.leeds.ac.uk> (aka <matthew@verisity.com>)
;;	Matthias Blume <blume@cs.princeton.edu> (aka <blume@kurims.kyoto-u.ac.jp>)
;;      (Stefan Monnier) monnier@cs.yale.edu
;;      Ian Zerny <ian@zerny.dk>
;; Maintainer: (Ian Zerny)
;; Keywords: DML
;; 0.01 (forked from sml-mode: 1.35)
;; 2012/04/01 00:00:00

;; This file is not part of GNU Emacs, but it is distributed under the
;; same conditions.

;; This program is free software; you can redistribute it and/or
;; modify it under the terms of the GNU General Public License as
;; published by the Free Software Foundation; either version 2, or (at
;; your option) any later version.

;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
;; General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING. If not, write to the
;; Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

;;; Commentary:

;;; HISTORY

;; Still under construction: History obscure, needs a biographer as
;; well as a M-x doctor. Change Log on request.

;; Hacked by Olin Shivers for comint from Lars Bo Nielsen's sml.el.

;; Hacked by Matthew Morley to incorporate Fritz Knabe's hilite and
;; font-lock patterns, some of Steven Gilmore's (reduced) easy-menus,
;; and numerous bugs and bug-fixes.

;; Hacked by Ian Zerny to replace (almost) all occurances of sml by
;; dml.  Hardcoded preprocessing by dmlc script when loading code.
;; Added DML keywords for syntax highlighting.

;;; DESCRIPTION

;; See accompanying info file in original sml-mode: sml-mode.info

;;; FOR YOUR .EMACS FILE

;; If dml-mode.el lives in some non-standard directory, you must tell
;; emacs where to get it. This may or may not be necessary:

;; (add-to-list 'load-path "~jones/lib/emacs/")

;; Then to access the commands autoload dml-mode with that command:

;; (load "dml-mode-startup")

;; dml-mode-hook is run whenever a new dml-mode buffer is created.

;; Finally, there are inferior-dml-{mode,load}-hooks -- see comments
;; in dml-proc.el. For much more information consult sml-mode's *info*
;; tree.

;;; Code:

(eval-when-compile (require 'cl))
(require 'dml-util)
(require 'dml-move)
(require 'dml-defs)
(condition-case nil (require 'skeleton) (error nil))

;;; VARIABLES CONTROLLING INDENTATION

(defcustom dml-indent-level 4
  "*Indentation of blocks in ML (see also `dml-structure-indent')."
  :group 'dml
  :type '(integer))

(defcustom dml-indent-args dml-indent-level
  "*Indentation of args placed on a separate line."
  :group 'dml
  :type '(integer))

;; (defvar dml-indent-align-args t
;;   "*Whether the arguments should be aligned.")

;; (defvar dml-case-indent nil
;;   "*How to indent case-of expressions.
;;     If t:   case expr                     If nil:   case expr of
;;               of exp1 => ...                            exp1 => ...
;;                | exp2 => ...                          | exp2 => ...

;; The first seems to be the standard in DML/NJ, but the second
;; seems nicer...")

(defcustom dml-electric-semi-mode nil
  "*If non-nil, `\;' will self insert, reindent the line, and do a newline.
If nil, just insert a `\;'.  (To insert while t, do: \\[quoted-insert] \;)."
  :group 'dml
  :type 'boolean)

(defcustom dml-rightalign-and t
  "If non-nil, right-align `and' with its leader.
If nil:					If t:
	datatype a = A				datatype a = A
	and b = B				     and b = B"
  :group 'dml
  :type 'boolean)

;;; OTHER GENERIC MODE VARIABLES

(defvar dml-mode-info "dml-mode"
  "*Where to find Info file for `dml-mode'.
The default assumes the info file \"dml-mode.info\" is on Emacs' info
directory path.  If it is not, either put the file on the standard path
or set the variable `dml-mode-info' to the exact location of this file

  (setq dml-mode-info \"/usr/me/lib/info/dml-mode\")

in your .emacs file. You can always set it interactively with the
set-variable command.")

(defvar dml-mode-hook nil
  "*Run upon entering `dml-mode'.
This is a good place to put your preferred key bindings.")

;;; CODE FOR DML-MODE

(defun dml-mode-info ()
  "Command to access the TeXinfo documentation for `dml-mode'.
See doc for the variable `dml-mode-info'."
  (interactive)
  (require 'info)
  (condition-case nil
      (info dml-mode-info)
    (error (progn
             (describe-variable 'dml-mode-info)
             (message "Can't find it... set this variable first!")))))


;;; Autoload functions -- no-doc is another idea cribbed from AucTeX!

(let ((dml-no-doc
       "This function is part of dml-proc, and has not yet been loaded.
Full documentation will be available after autoloading the function."))

  (autoload 'dml-compile	"dml-proc"   dml-no-doc t)
  (autoload 'dml-load-file	"dml-proc"   dml-no-doc t)
  (autoload 'switch-to-dml	"dml-proc"   dml-no-doc t)
  (autoload 'dml-send-region	"dml-proc"   dml-no-doc t)
  (autoload 'dml-send-buffer	"dml-proc"   dml-no-doc t))

;; font-lock setup

(defconst dml-keywords-regexp
  (dml-syms-re "abstraction" "abstype" "and" "andalso" "as" "before" "case"
	       "datatype" "else" "end" "eqtype" "exception" "do" "fn"
	       "fun" "functor" "handle" "if" "in" "include" "infix"
	       "infixr" "let" "local" "nonfix" "of" "op" "open" "orelse"
	       "overload" "raise" "rec" "sharing" "sig" "signature"
	       "struct" "structure" "then" "type" "val" "where" "while"
	       "with" "withtype" "o"
	       "FUN" "NUF" "FN" "NF" "IF" "THEN" "ELSE" "MAIN"
	       "TRUE" "FALSE" "BOOL" "NUM" "STR" "LIST"
	       "asBOOL" "asNUM" "asSTR" "asLIST"
	       "isBOOL" "isNUM" "isSTR" "isLIST"
	       "PRINT" "PRINTLN")
  "A regexp that matches any and all keywords of DML.")

(defconst dml-tyvarseq-re
  "\\(\\('+\\(\\sw\\|\\s_\\)+\\|(\\([,']\\|\\sw\\|\\s_\\|\\s-\\)+)\\)\\s-+\\)?")

;;; Font-lock settings ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defcustom dml-font-lock-symbols nil
  "Display \\ and -> and such using symbols in fonts.
This may sound like a neat trick, but be extra careful: it changes the
alignment and can thus lead to nasty surprises w.r.t layout.
If t, try to use whichever font is available.  Otherwise you can
set it to a particular font of your preference among `japanese-jisx0208'
and `unicode'."
  :type '(choice (const nil)
	         (const t)
	         (const unicode)
	         (const japanese-jisx0208)))

(defconst dml-font-lock-symbols-alist
  (append
   ;; The symbols can come from a JIS0208 font.
   (and (fboundp 'make-char) (fboundp 'charsetp) (charsetp 'japanese-jisx0208)
	(memq dml-font-lock-symbols '(t japanese-jisx0208))
	(list (cons "fn" (make-char 'japanese-jisx0208 38 75))
	      (cons "andalso" (make-char 'japanese-jisx0208 34 74))
	      (cons "orelse" (make-char 'japanese-jisx0208 34 75))
	      ;; (cons "as" (make-char 'japanese-jisx0208 34 97))
	      (cons "not" (make-char 'japanese-jisx0208 34 76))
	      (cons "div" (make-char 'japanese-jisx0208 33 96))
	      ;; (cons "*" (make-char 'japanese-jisx0208 33 95))
	      (cons "->" (make-char 'japanese-jisx0208 34 42))
	      (cons "=>" (make-char 'japanese-jisx0208 34 77))
	      (cons "<-" (make-char 'japanese-jisx0208 34 43))
	      (cons "<>" (make-char 'japanese-jisx0208 33 98))
	      (cons ">=" (make-char 'japanese-jisx0208 33 102))
	      (cons "<=" (make-char 'japanese-jisx0208 33 101))
	      (cons "..." (make-char 'japanese-jisx0208 33 68))
	      ;; Some greek letters for type parameters.
	      (cons "'a" (make-char 'japanese-jisx0208 38 65))
	      (cons "'b" (make-char 'japanese-jisx0208 38 66))
	      (cons "'c" (make-char 'japanese-jisx0208 38 67))
	      (cons "'d" (make-char 'japanese-jisx0208 38 68))
	      ))
   ;; Or a unicode font.
   (and (fboundp 'decode-char)
	(memq dml-font-lock-symbols '(t unicode))
	(list (cons "fn" (decode-char 'ucs 955))
	      (cons "andalso" (decode-char 'ucs 8896))
	      (cons "orelse" (decode-char 'ucs 8897))
	      ;; (cons "as" (decode-char 'ucs 8801))
	      (cons "not" (decode-char 'ucs 160))
	      (cons "div" (decode-char 'ucs 247))
	      (cons "*" (decode-char 'ucs 215))
	      (cons "o"  (decode-char 'ucs 9675))
	      (cons "->" (decode-char 'ucs 8594))
	      (cons "=>" (decode-char 'ucs 8658))
	      (cons "<-" (decode-char 'ucs 8592))
	      (cons "<>" (decode-char 'ucs 8800))
	      (cons ">=" (decode-char 'ucs 8805))
	      (cons "<=" (decode-char 'ucs 8804))
	      (cons "..." (decode-char 'ucs 8943))
	      ;; (cons "::" (decode-char 'ucs 8759))
	      ;; Some greek letters for type parameters.
	      (cons "'a" (decode-char 'ucs 945))
	      (cons "'b" (decode-char 'ucs 946))
	      (cons "'c" (decode-char 'ucs 947))
	      (cons "'d" (decode-char 'ucs 948))
	      ))))

(defun dml-font-lock-compose-symbol (alist)
  "Compose a sequence of ascii chars into a symbol.
Regexp match data 0 points to the chars."
  ;; Check that the chars should really be composed into a symbol.
  (let* ((start (match-beginning 0))
	 (end (match-end 0))
	 (syntaxes (if (eq (char-syntax (char-after start)) ?w)
		       '(?w) '(?. ?\\))))
    (if (or (memq (char-syntax (or (char-before start) ?\ )) syntaxes)
	    (memq (char-syntax (or (char-after end) ?\ )) syntaxes)
	    (memq (get-text-property start 'face)
		  '(font-lock-doc-face font-lock-string-face
		    font-lock-comment-face)))
	;; No composition for you.  Let's actually remove any composition
	;; we may have added earlier and which is now incorrect.
	(remove-text-properties start end '(composition))
      ;; That's a symbol alright, so add the composition.
      (compose-region start end (cdr (assoc (match-string 0) alist)))))
  ;; Return nil because we're not adding any face property.
  nil)

(defun dml-font-lock-symbols-keywords ()
  (when (fboundp 'compose-region)
    (let ((alist nil))
      (dolist (x dml-font-lock-symbols-alist)
	(when (and (if (fboundp 'char-displayable-p)
		       (char-displayable-p (cdr x))
		     t)
		   (not (assoc (car x) alist)))	;Not yet in alist.
	  (push x alist)))
      (when alist
	`((,(regexp-opt (mapcar 'car alist) t)
	   (0 (dml-font-lock-compose-symbol ',alist))))))))

;; The font lock regular expressions.

(defconst dml-font-lock-keywords
  `(;;(dml-font-comments-and-strings)
    (,(concat "\\<\\(fun\\|and\\)\\s-+" dml-tyvarseq-re "\\(\\sw+\\)\\s-+[^ \t\n=]")
     (1 font-lock-keyword-face)
     (6 font-lock-function-name-face))
    (,(concat "\\<\\(\\(data\\|abs\\|with\\|eq\\)?type\\)\\s-+" dml-tyvarseq-re "\\(\\sw+\\)")
     (1 font-lock-keyword-face)
     (7 font-lock-type-def-face))
    ("\\<\\(val\\)\\s-+\\(\\sw+\\>\\s-*\\)?\\(\\sw+\\)\\s-*[=:]"
     (1 font-lock-keyword-face)
     ;;(6 font-lock-variable-def-face nil t)
     (3 font-lock-variable-name-face))
    ("\\<\\(structure\\|functor\\|abstraction\\)\\s-+\\(\\sw+\\)"
     (1 font-lock-keyword-face)
     (2 font-lock-module-def-face))
    ("\\<\\(signature\\)\\s-+\\(\\sw+\\)"
     (1 font-lock-keyword-face)
     (2 font-lock-interface-def-face))
    
    (,dml-keywords-regexp . font-lock-keyword-face)
    ,@(dml-font-lock-symbols-keywords))
  "Regexps matching standard DML keywords.")

(defface font-lock-type-def-face
  '((t (:bold t)))
  "Font Lock mode face used to highlight type definitions."
  :group 'font-lock-highlighting-faces)
(defvar font-lock-type-def-face 'font-lock-type-def-face
  "Face name to use for type definitions.")

(defface font-lock-module-def-face
  '((t (:bold t)))
  "Font Lock mode face used to highlight module definitions."
  :group 'font-lock-highlighting-faces)
(defvar font-lock-module-def-face 'font-lock-module-def-face
  "Face name to use for module definitions.")

(defface font-lock-interface-def-face
  '((t (:bold t)))
  "Font Lock mode face used to highlight interface definitions."
  :group 'font-lock-highlighting-faces)
(defvar font-lock-interface-def-face 'font-lock-interface-def-face
  "Face name to use for interface definitions.")

;;
;; Code to handle nested comments and unusual string escape sequences
;;

(defsyntax dml-syntax-prop-table
  '((?\\ . ".") (?* . "."))
  "Syntax table for text-properties")

;; For Emacsen that have no built-in support for nested comments
(defun dml-get-depth-st ()
  (save-excursion
    (let* ((disp (if (eq (char-before) ?\)) (progn (backward-char) -1) nil))
	   (_ (backward-char))
	   (disp (if (eq (char-before) ?\() (progn (backward-char) 0) disp))
	   (pt (point)))
      (when disp
	(let* ((depth
		(save-match-data
		  (if (re-search-backward "\\*)\\|(\\*" nil t)
		      (+ (or (get-char-property (point) 'comment-depth) 0)
			 (case (char-after) (?\( 1) (?* 0))
			 disp)
		    0)))
	       (depth (if (> depth 0) depth)))
	  (put-text-property pt (1+ pt) 'comment-depth depth)
	  (when depth dml-syntax-prop-table))))))

(defconst dml-font-lock-syntactic-keywords
  `(("^\\s-*\\(\\\\\\)" (1 ',dml-syntax-prop-table))
    ,@(unless dml-builtin-nested-comments-flag
	'(("(?\\(\\*\\))?" (1 (dml-get-depth-st)))))))

(defconst dml-font-lock-defaults
  '(dml-font-lock-keywords nil nil ((?_ . "w") (?' . "w")) nil
    (font-lock-syntactic-keywords . dml-font-lock-syntactic-keywords)))

;;;;
;;;; Imenu support
;;;;

(defvar dml-imenu-regexp
  (concat "^[ \t]*\\(let[ \t]+\\)?"
	  (regexp-opt (append dml-module-head-syms
			      '("and" "fun" "datatype" "abstype" "type")) t)
	  "\\>"))

(defun dml-imenu-create-index ()
  (let (alist)
    (goto-char (point-max))
    (while (re-search-backward dml-imenu-regexp nil t)
      (save-excursion
	(let ((kind (match-string 2))
	      (column (progn (goto-char (match-beginning 2)) (current-column)))
	      (location
	       (progn (goto-char (match-end 0))
		      (dml-forward-spaces)
		      (when (looking-at dml-tyvarseq-re)
			(goto-char (match-end 0)))
		      (point)))
	      (name (dml-forward-sym)))
	  ;; Eliminate trivial renamings.
	  (when (or (not (member kind '("structure" "signature")))
		    (progn (search-forward "=")
			   (dml-forward-spaces)
			   (looking-at "sig\\|struct")))
	    (push (cons (concat (make-string (/ column 2) ?\ ) name) location)
		  alist)))))
    alist))

;;; MORE CODE FOR DML-MODE

;;;###autoload (add-to-list 'load-path (file-name-directory load-file-name))
;;;###autoload
(add-to-list 'auto-mode-alist '("\\.d\\(ml\\|ig\\)\\'" . dml-mode))

;;;###autoload
(define-derived-mode dml-mode fundamental-mode "DML"
  "\\<dml-mode-map>Major mode for editing ML code.
This mode runs `dml-mode-hook' just before exiting.
\\{dml-mode-map}"
  (set (make-local-variable 'font-lock-defaults) dml-font-lock-defaults)
  (set (make-local-variable 'outline-regexp) dml-outline-regexp)
  (set (make-local-variable 'imenu-create-index-function)
       'dml-imenu-create-index)
  (set (make-local-variable 'add-log-current-defun-function)
       'dml-current-fun-name)
  ;; Treat paragraph-separators in comments as paragraph-separators.
  (set (make-local-variable 'paragraph-separate)
       (concat "\\([ \t]*\\*)?\\)?\\(" paragraph-separate "\\)"))
  (set (make-local-variable 'require-final-newline) t)
  ;; forward-sexp-function is an experimental variable in my hacked Emacs.
  (set (make-local-variable 'forward-sexp-function) 'dml-user-forward-sexp)
  ;; For XEmacs
  (easy-menu-add dml-mode-menu)
  ;; Compatibility.  FIXME: we should use `-' in Emacs-CVS.
  (unless (boundp 'skeleton-positions) (set (make-local-variable '@) nil))
  (dml-mode-variables))

(defun dml-mode-variables ()
  (set-syntax-table dml-mode-syntax-table)
  (setq local-abbrev-table dml-mode-abbrev-table)
  ;; A paragraph is separated by blank lines or ^L only.
  
  (set (make-local-variable 'indent-line-function) 'dml-indent-line)
  (set (make-local-variable 'comment-start) "(* ")
  (set (make-local-variable 'comment-end) " *)")
  (set (make-local-variable 'comment-nested) t)
  ;;(set (make-local-variable 'block-comment-start) "* ")
  ;;(set (make-local-variable 'block-comment-end) "")
  ;; (set (make-local-variable 'comment-column) 40)
  (set (make-local-variable 'comment-start-skip) "(\\*+\\s-*"))

(defun dml-funname-of-and ()
  "Name of the function this `and' defines, or nil if not a function.
Point has to be right after the `and' symbol and is not preserved."
  (dml-forward-spaces)
  (if (looking-at dml-tyvarseq-re) (goto-char (match-end 0)))
  (let ((sym (dml-forward-sym)))
    (dml-forward-spaces)
    (unless (or (member sym '(nil "d="))
		(member (dml-forward-sym) '("d=")))
      sym)))

(defun dml-electric-pipe ()
  "Insert a \"|\".
Depending on the context insert the name of function, a \"=>\" etc."
  (interactive)
  (dml-with-ist
   (unless (save-excursion (skip-chars-backward "\t ") (bolp)) (insert "\n"))
   (insert "| ")
   (let ((text
	  (save-excursion
	    (backward-char 2)		;back over the just inserted "| "
	    (let ((sym (dml-find-matching-starter dml-pipeheads
						  (dml-op-prec "|" 'back))))
	      (dml-forward-sym)
	      (dml-forward-spaces)
	      (cond
	       ((string= sym "|")
		(let ((f (dml-forward-sym)))
		  (dml-find-forward "\\(=>\\|=\\||\\)\\S.")
		  (cond
		   ((looking-at "|") "") ;probably a datatype
		   ((looking-at "=>") " => ") ;`case', or `fn' or `handle'
		   ((looking-at "=") (concat f "  = "))))) ;a function
	       ((string= sym "and")
		;; could be a datatype or a function
		(setq sym (dml-funname-of-and))
		(if sym (concat sym "  = ") ""))
	       ;; trivial cases
	       ((string= sym "fun")
		(while (and (setq sym (dml-forward-sym))
			    (string-match "^'" sym))
		  (dml-forward-spaces))
		(concat sym "  = "))
	       ((member sym '("case" "handle" "fn" "of")) " => ")
	       ;;((member sym '("abstype" "datatype")) "")
	       (t ""))))))

     (insert text)
     (indent-according-to-mode)
     (beginning-of-line)
     (skip-chars-forward "\t |")
     (skip-syntax-forward "w")
     (skip-chars-forward "\t ")
     (when (eq ?= (char-after)) (backward-char)))))

(defun dml-electric-semi ()
  "Insert a \;.
If variable `dml-electric-semi-mode' is t, indent the current line, insert
a newline, and indent."
  (interactive)
  (insert "\;")
  (if dml-electric-semi-mode
      (reindent-then-newline-and-indent)))

;;; INDENTATION !!!

(defun dml-mark-function ()
  "Synonym for `mark-paragraph' -- sorry.
If anyone has a good algorithm for this..."
  (interactive)
  (mark-paragraph))

(defun dml-indent-line ()
  "Indent current line of ML code."
  (interactive)
  (let ((savep (> (current-column) (current-indentation)))
	(indent (max (or (ignore-errors (dml-calculate-indentation)) 0) 0)))
    (if savep
	(save-excursion (indent-line-to indent))
      (indent-line-to indent))))

(defun dml-back-to-outer-indent ()
  "Unindents to the next outer level of indentation."
  (interactive)
  (save-excursion
    (beginning-of-line)
    (skip-chars-forward "\t ")
    (let ((start-column (current-column))
          (indent (current-column)))
      (if (> start-column 0)
          (progn
            (save-excursion
              (while (>= indent start-column)
                (if (re-search-backward "^[^\n]" nil t)
                    (setq indent (current-indentation))
                  (setq indent 0))))
            (backward-delete-char-untabify (- start-column indent)))))))

(defun dml-find-comment-indent ()
  (save-excursion
    (let ((depth 1))
      (while (> depth 0)
	(if (re-search-backward "(\\*\\|\\*)" nil t)
	    (cond
	     ;; FIXME: That's just a stop-gap.
	     ((eq (get-text-property (point) 'face) 'font-lock-string-face))
	     ((looking-at "*)") (incf depth))
	     ((looking-at comment-start-skip) (decf depth)))
	  (setq depth -1)))
      (if (= depth 0)
	  (1+ (current-column))
	nil))))

(defun dml-calculate-indentation ()
  (save-excursion
    (beginning-of-line) (skip-chars-forward "\t ")
    (dml-with-ist
     ;; Indentation for comments alone on a line, matches the
     ;; proper indentation of the next line.
     (when (looking-at "(\\*") (dml-forward-spaces))
     (let (data
	   (sym (save-excursion (dml-forward-sym))))
       (or
	;; Allow the user to override the indentation.
	(when (looking-at (concat ".*" (regexp-quote comment-start)
				  "[ \t]*fixindent[ \t]*"
				  (regexp-quote comment-end)))
	  (current-indentation))

	;; Continued comment.
	(and (looking-at "\\*") (dml-find-comment-indent))

	;; Continued string ? (Added 890113 lbn)
	(and (looking-at "\\\\")
	     (save-excursion
	       (if (save-excursion (previous-line 1)
				   (beginning-of-line)
				   (looking-at "[\t ]*\\\\"))
		   (progn (previous-line 1) (current-indentation))
		 (if (re-search-backward "[^\\\\]\"" nil t)
		     (1+ (current-column))
		   0))))

	;; Closing parens.  Could be handled below with `dml-indent-relative'?
	(and (looking-at "\\s)")
	     (save-excursion
	       (skip-syntax-forward ")")
	       (backward-sexp 1)
	       (if (dml-dangling-sym)
		   (dml-indent-default 'noindent)
		 (current-column))))

	(and (setq data (assoc sym dml-close-paren))
	     (dml-indent-relative sym data))

	(and (member sym dml-starters-syms)
	     (dml-indent-starter sym))

	(and (string= sym "|") (dml-indent-pipe))

	(dml-indent-arg)
	(dml-indent-default))))))

(defsubst dml-bolp ()
  (save-excursion (skip-chars-backward " \t|") (bolp)))

(defun dml-indent-starter (orig-sym)
  "Return the indentation to use for a symbol in `dml-starters-syms'.
Point should be just before the symbol ORIG-SYM and is not preserved."
  (let ((sym (unless (save-excursion (dml-backward-arg))
	       (dml-backward-spaces)
	       (dml-backward-sym))))
    (if (member sym '(";" "d=")) (setq sym nil))
    (if sym (dml-get-sym-indent sym)
      ;; FIXME: this can take a *long* time !!
      (setq sym (dml-find-matching-starter dml-starters-syms))
      ;; Don't align with `and' because it might be specially indented.
      (if (and (or (equal orig-sym "and") (not (equal sym "and")))
	       (dml-bolp))
	  (+ (current-column)
	     (if (and dml-rightalign-and (equal orig-sym "and"))
		 (- (length sym) 3) 0))
	(dml-indent-starter orig-sym)))))

(defun dml-indent-relative (sym data)
  (save-excursion
    (dml-forward-sym) (dml-backward-sexp nil)
    (unless (second data) (dml-backward-spaces) (dml-backward-sym))
    (+ (or (cdr (assoc sym dml-symbol-indent)) 0)
       (dml-delegated-indent))))

(defun dml-indent-pipe ()
  (let ((sym (dml-find-matching-starter dml-pipeheads
					(dml-op-prec "|" 'back))))
    (when sym
      (if (string= sym "|")
	  (if (dml-bolp) (current-column) (dml-indent-pipe))
	(let ((pipe-indent (or (cdr (assoc "|" dml-symbol-indent)) -2)))
	  (when (or (member sym '("datatype" "abstype"))
		    (and (equal sym "and")
			 (save-excursion
			   (forward-word 1)
			   (not (dml-funname-of-and)))))
	    (re-search-forward "="))
	  (dml-forward-sym)
	  (dml-forward-spaces)
	  (+ pipe-indent (current-column)))))))

(defun dml-find-forward (re)
  (dml-forward-spaces)
  (while (and (not (looking-at re))
	      (progn
		(or (ignore-errors (forward-sexp 1) t) (forward-char 1))
		(dml-forward-spaces)
		(not (looking-at re))))))

(defun dml-indent-arg ()
  (and (save-excursion (ignore-errors (dml-forward-arg)))
       ;;(not (looking-at dml-not-arg-re))
       ;; looks like a function or an argument
       (dml-move-if (dml-backward-arg))
       ;; an argument
       (if (save-excursion (not (dml-backward-arg)))
	   ;; a first argument
	   (+ (current-column) dml-indent-args)
	 ;; not a first arg
	 (while (and (/= (current-column) (current-indentation))
		     (dml-move-if (dml-backward-arg))))
	 (unless (save-excursion (dml-backward-arg))
	   ;; all earlier args are on the same line
	   (dml-forward-arg) (dml-forward-spaces))
	 (current-column))))

(defun dml-get-indent (data sym)
  (let (d)
    (cond
     ((not (listp data)) data)
     ((setq d (member sym data)) (cadr d))
     ((and (consp data) (not (stringp (car data)))) (car data))
     (t dml-indent-level))))

(defun dml-dangling-sym ()
  "Non-nil if the symbol after point is dangling.
The symbol can be an DML symbol or an open-paren. \"Dangling\" means that
it is not on its own line but is the last element on that line."
  (save-excursion
    (and (not (dml-bolp))
	 (< (dml-point-after (end-of-line))
	    (dml-point-after (or (dml-forward-sym) (skip-syntax-forward "("))
			     (dml-forward-spaces))))))

(defun dml-delegated-indent ()
  (if (dml-dangling-sym)
      (dml-indent-default 'noindent)
    (dml-move-if (backward-word 1)
		 (looking-at dml-agglomerate-re))
    (current-column)))

(defun dml-get-sym-indent (sym &optional style)
  "Find the indentation for the SYM we're `looking-at'.
If indentation is delegated, point will move to the start of the parent.
Optional argument STYLE is currently ignored."
  (assert (equal sym (save-excursion (dml-forward-sym))))
  (save-excursion
    (let ((delegate (and (not (equal sym "end")) (assoc sym dml-close-paren)))
	  (head-sym sym))
      (when (and delegate (not (eval (third delegate))))
	;;(dml-find-match-backward sym delegate)
	(dml-forward-sym) (dml-backward-sexp nil)
	(setq head-sym
	      (if (second delegate)
		  (save-excursion (dml-forward-sym))
		(dml-backward-spaces) (dml-backward-sym))))

      (let ((idata (assoc head-sym dml-indent-rule)))
	(when idata
	  ;;(if (or style (not delegate))
	  ;; normal indentation
	  (let ((indent (dml-get-indent (cdr idata) sym)))
	    (when indent (+ (dml-delegated-indent) indent)))
	  ;; delgate indentation to the parent
	  ;;(dml-forward-sym) (dml-backward-sexp nil)
	  ;;(let* ((parent-sym (save-excursion (dml-forward-sym)))
	  ;;     (parent-indent (cdr (assoc parent-sym dml-indent-starters))))
	  ;; check the special rules
	  ;;(+ (dml-delegated-indent)
	  ;; (or (dml-get-indent (cdr indent-data) 1 'strict)
	  ;; (dml-get-indent (cdr parent-indent) 1 'strict)
	  ;; (dml-get-indent (cdr indent-data) 0)
	  ;; (dml-get-indent (cdr parent-indent) 0))))))))
	  )))))

(defun dml-indent-default (&optional noindent)
  (let* ((sym-after (save-excursion (dml-forward-sym)))
	 (_ (dml-backward-spaces))
	 (sym-before (dml-backward-sym))
	 (sym-indent (and sym-before (dml-get-sym-indent sym-before)))
	 (indent-after (or (cdr (assoc sym-after dml-symbol-indent)) 0)))
    (when (equal sym-before "end")
      ;; I don't understand what's really happening here, but when
      ;; it's `end' clearly, we need to do something special.
      (forward-word 1)
      (setq sym-before nil sym-indent nil))
    (cond
     (sym-indent
      ;; the previous sym is an indentation introducer: follow the rule
      (if noindent
	  ;;(current-column)
	  sym-indent
	(+ sym-indent indent-after)))
     ;; If we're just after a hanging open paren.
     ((and (eq (char-syntax (preceding-char)) ?\()
	   (save-excursion (backward-char) (dml-dangling-sym)))
      (backward-char)
      (dml-indent-default))
     (t
      ;; default-default
      (let* ((prec-after (dml-op-prec sym-after 'back))
	     (prec (or (dml-op-prec sym-before 'back) prec-after 100)))
	;; go back until you hit a symbol that has a lower prec than the
	;; "current one", or until you backed over a sym that has the same prec
	;; but is at the beginning of a line.
	(while (and (not (dml-bolp))
		    (while (dml-move-if (dml-backward-sexp (1- prec))))
		    (not (dml-bolp)))
	  (while (dml-move-if (dml-backward-sexp prec))))
	(if noindent
	    ;; the `noindent' case does back over an introductory symbol
	    ;; such as `fun', ...
	    (progn
	      (dml-move-if
	       (dml-backward-spaces)
	       (member (dml-backward-sym) dml-starters-syms))
	      (current-column))
	  ;; Use `indent-after' for cases such as when , or ; should be
	  ;; outdented so that their following terms are aligned.
	  (+ (if (progn
		   (if (equal sym-after ";")
		       (dml-move-if
			(dml-backward-spaces)
			(member (dml-backward-sym) dml-starters-syms)))
		   (and sym-after (not (looking-at sym-after))))
		 indent-after 0)
	     (current-column))))))))


;; maybe `|' should be set to word-syntax in our temp syntax table ?
(defun dml-current-indentation ()
  (save-excursion
    (beginning-of-line)
    (skip-chars-forward " \t|")
    (current-column)))


(defun dml-find-matching-starter (syms &optional prec)
  (let (sym)
    (ignore-errors
      (while
	  (progn (dml-backward-sexp prec)
		 (setq sym (save-excursion (dml-forward-sym)))
		 (not (or (member sym syms) (bobp)))))
      (if (member sym syms) sym))))

(defun dml-skip-siblings ()
  (while (and (not (bobp)) (dml-backward-arg))
    (dml-find-matching-starter dml-starters-syms))
  (when (looking-at "in\\>\\|local\\>")
    ;;skip over `local...in' and continue
    (forward-word 1)
    (dml-backward-sexp nil)
    (dml-skip-siblings)))

(defun dml-beginning-of-defun ()
  (let ((sym (dml-find-matching-starter dml-starters-syms)))
    (if (member sym '("fun" "and" "functor" "signature" "structure"
		      "abstraction" "datatype" "abstype"))
	(save-excursion (dml-forward-sym) (dml-forward-spaces)
			(dml-forward-sym))
      ;; We're inside a "non function declaration": let's skip all other
      ;; declarations that we find at the same level and try again.
      (dml-skip-siblings)
      ;; Obviously, let's not try again if we're at bobp.
      (unless (bobp) (dml-beginning-of-defun)))))

(defcustom dml-max-name-components 3
  "Maximum number of components to use for the current function name."
  :group 'dml
  :type 'integer)

(defun dml-current-fun-name ()
  (save-excursion
    (let ((count dml-max-name-components)
	  fullname name)
      (end-of-line)
      (while (and (> count 0)
		  (setq name (dml-beginning-of-defun)))
	(decf count)
	(setq fullname (if fullname (concat name "." fullname) name))
	;; Skip all other declarations that we find at the same level.
	(dml-skip-siblings))
      fullname)))


;;; INSERTING PROFORMAS (COMMON DML-FORMS)

(defvar dml-forms-alist nil
  "*Alist of code templates.
You can extend this alist to your heart's content.  For each additional
template NAME in the list, declare a keyboard macro or function (or
interactive command) called 'dml-form-NAME'.
If 'dml-form-NAME' is a function it takes no arguments and should
insert the template at point\; if this is a command it may accept any
sensible interactive call arguments\; keyboard macros can't take
arguments at all.  Apropos keyboard macros, see `name-last-kbd-macro'
and `dml-addto-forms-alist'.
`dml-forms-alist' understands let, local, case, abstype, datatype,
signature, structure, and functor by default.")

(defmacro dml-def-skeleton (name interactor &rest elements)
  (when (fboundp 'define-skeleton)
    (let ((fsym (intern (concat "dml-form-" name))))
      `(progn
	 (add-to-list 'dml-forms-alist ',(cons name fsym))
	 (condition-case err
	     ;; Try to use the new `system' flag.
	     (define-abbrev dml-mode-abbrev-table ,name "" ',fsym nil 'system)
	   (wrong-number-of-arguments
	    (define-abbrev dml-mode-abbrev-table ,name "" ',fsym)))
	 (define-skeleton ,fsym
	   ,(format "DML-mode skeleton for `%s..' expressions" name)
	   ,interactor
	   ,(concat name " ") >
	   ,@elements)))))
(put 'dml-def-skeleton 'lisp-indent-function 2)

(dml-def-skeleton "let" nil
  @ "\nin " > _ "\nend" >)

(dml-def-skeleton "if" nil
  @ " then " > _ "\nelse " > _)

(dml-def-skeleton "local" nil
  @ "\nin" > _ "\nend" >)

(dml-def-skeleton "case" "Case expr: "
  str "\nof " > _ " => ")

(dml-def-skeleton "signature" "Signature name: "
  str " =\nsig" > "\n" > _ "\nend" >)

(dml-def-skeleton "structure" "Structure name: "
  str " =\nstruct" > "\n" > _ "\nend" >)

(dml-def-skeleton "functor" "Functor name: "
  str " () : =\nstruct" > "\n" > _ "\nend" >)

(dml-def-skeleton "datatype" "Datatype name and type params: "
  str " =" \n)

(dml-def-skeleton "abstype" "Abstype name and type params: "
  str " =" \n _ "\nwith" > "\nend" >)

;;

(dml-def-skeleton "struct" nil
  _ "\nend" >)

(dml-def-skeleton "sig" nil
  _ "\nend" >)

(dml-def-skeleton "val" nil
  @ " = " > _)

(dml-def-skeleton "fn" nil
  @ " =>" > _)

(dml-def-skeleton "fun" nil
  @ " =" > _)

;;

(defun dml-forms-menu (menu)
  (mapcar (lambda (x) (vector (car x) (cdr x) t))
	  dml-forms-alist))

(defvar dml-last-form "let")

(defun dml-electric-space ()
  "Expand a symbol into an DML form, or just insert a space.
If the point directly precedes a symbol for which an DML form exists,
the corresponding form is inserted."
  (interactive)
  (let ((abbrev-mode (not abbrev-mode))
	(last-command-char ?\ )
	;; Bind `this-command' to fool skeleton's special abbrev handling.
	(this-command 'self-insert-command))
    (call-interactively 'self-insert-command)))

(defun dml-insert-form (name newline)
  "Interactive short-cut to insert the NAME common ML form.
If a prefix argument is given insert a NEWLINE and indent first, or
just move to the proper indentation if the line is blank\; otherwise
insert at point (which forces indentation to current column).

The default form to insert is 'whatever you inserted last time'
\(just hit return when prompted\)\; otherwise the command reads with
completion from `dml-forms-alist'."
  (interactive
   (list (completing-read
	  (format "Form to insert: (default %s) " dml-last-form)
	  dml-forms-alist nil t nil)
	 current-prefix-arg))
  ;; default is whatever the last insert was...
  (if (string= name "") (setq name dml-last-form) (setq dml-last-form name))
  (unless (or (not newline)
	      (save-excursion (beginning-of-line) (looking-at "\\s-*$")))
    (insert "\n"))
  (unless (/= ?w (char-syntax (preceding-char))) (insert " "))
  (let ((f (cdr (assoc name dml-forms-alist))))
    (cond
     ((commandp f) (command-execute f))
     (f (funcall f))
     (t (error "Undefined form: %s" name)))))

;; See also macros.el in emacs lisp dir.

(defun dml-addto-forms-alist (name)
  "Assign a name to the last keyboard macro defined.
Argument NAME is transmogrified to dml-form-NAME which is the symbol
actually defined.

The symbol's function definition becomes the keyboard macro string.

If that works, NAME is added to `dml-forms-alist' so you'll be able to
reinvoke the macro through \\[dml-insert-form].  You might want to save
the macro to use in a later editing session -- see `insert-kbd-macro'
and add these macros to your .emacs file.

See also `edit-kbd-macro' which is bound to \\[edit-kbd-macro]."
  (interactive "sName for last kbd macro (\"dml-form-\" will be added): ")
  (when (string= name "") (error "No command name given"))
  (let ((fsym (intern (concat "dml-form-" name))))
    (name-last-kbd-macro fsym)
    (message "Macro bound to %s" fsym)
    (add-to-list 'dml-forms-alist (cons name fsym))))

;;;;
;;;;  DML/NJ's Compilation Manager support
;;;;

(defvar dml-cm-mode-syntax-table dml-mode-syntax-table)
(defvar dml-cm-font-lock-keywords
 `(,(concat "\\<" (regexp-opt '("library" "group" "is" "structure"
				"functor" "signature" "funsig") t)
	    "\\>")))
;;;###autoload
(add-to-list 'completion-ignored-extensions "CM/")
(add-to-list 'completion-ignored-extensions ".cm/")
;;;###autoload
(add-to-list 'auto-mode-alist '("\\.cm\\'" . dml-cm-mode))
;;;###autoload
(define-derived-mode dml-cm-mode fundamental-mode "DML-CM"
  "Major mode for DML/NJ's Compilation Manager configuration files."
  (local-set-key "\C-c\C-c" 'dml-compile)
  (set (make-local-variable 'font-lock-defaults)
       '(dml-cm-font-lock-keywords nil t nil nil)))

;;;;
;;;; ML-Lex support
;;;;

(defvar dml-lex-font-lock-keywords
  (append
   '(("^%\\sw+" . font-lock-builtin-face)
     ("^%%" . font-lock-module-def-face))
   dml-font-lock-keywords))
(defconst dml-lex-font-lock-defaults
  (cons 'dml-lex-font-lock-keywords (cdr dml-font-lock-defaults)))

;;;###autoload
(define-derived-mode dml-lex-mode dml-mode "DML-Lex"
  "Major Mode for editing ML-Lex files."
  (set (make-local-variable 'font-lock-defaults) dml-lex-font-lock-defaults))

;;;;
;;;; ML-Yacc support
;;;;

(defface dml-yacc-bnf-face
  '((t (:foreground "darkgreen")))
  "Face used to highlight (non)terminals in `dml-yacc-mode'.")
(defvar dml-yacc-bnf-face 'dml-yacc-bnf-face)

(defcustom dml-yacc-indent-action 16
  "Indentation column of the opening paren of actions."
  :group 'dml
  :type 'integer)

(defcustom dml-yacc-indent-pipe nil
  "Indentation column of the pipe char in the BNF.
If nil, align it with `:' or with previous cases."
  :group 'dml
  :type 'integer)

(defcustom dml-yacc-indent-term nil
  "Indentation column of the (non)term part.
If nil, align it with previous cases."
  :group 'dml
  :type 'integer)

(defvar dml-yacc-font-lock-keywords
  (cons '("^\\(\\sw+\\s-*:\\|\\s-*|\\)\\(\\s-*\\sw+\\)*\\s-*\\(\\(%\\sw+\\)\\s-+\\sw+\\|\\)"
	  (0 (save-excursion
	       (save-match-data
		 (goto-char (match-beginning 0))
		 (unless (or (re-search-forward "\\<of\\>" (match-end 0) 'move)
			     (progn (dml-forward-spaces)
				    (not (looking-at "("))))
		   dml-yacc-bnf-face))))
	  (4 font-lock-builtin-face t t))
	dml-lex-font-lock-keywords))
(defconst dml-yacc-font-lock-defaults
  (cons 'dml-yacc-font-lock-keywords (cdr dml-font-lock-defaults)))

(defun dml-yacc-indent-line ()
  "Indent current line of ML-Yacc code."
  (let ((savep (> (current-column) (current-indentation)))
	(indent (max (or (ignore-errors (dml-yacc-indentation)) 0) 0)))
    (if savep
	(save-excursion (indent-line-to indent))
      (indent-line-to indent))))

(defun dml-yacc-indentation ()
  (save-excursion
    (back-to-indentation)
    (or (and (looking-at "%\\|\\(\\sw\\|\\s_\\)+\\s-*:") 0)
	(when (save-excursion
		(condition-case nil (progn (up-list -1) nil) (scan-error t)))
	  ;; We're outside an action.
	  (cond
	   ;; Special handling of indentation inside %term and %nonterm
	   ((save-excursion
	      (and (re-search-backward "^%\\(\\sw+\\)" nil t)
		   (member (match-string 1) '("term" "nonterm"))))
	    (if (numberp dml-yacc-indent-term) dml-yacc-indent-term
	      (let ((offset (if (looking-at "|") -2 0)))
		(forward-line -1)
		(looking-at "\\s-*\\(%\\sw*\\||\\)?\\s-*")
		(goto-char (match-end 0))
		(+ offset (current-column)))))
	   ((looking-at "(") dml-yacc-indent-action)
	   ((looking-at "|")
	    (if (numberp dml-yacc-indent-pipe) dml-yacc-indent-pipe
	      (backward-sexp 1)
	      (while (progn (dml-backward-spaces)
			    (/= 0 (skip-syntax-backward "w_"))))
	      (dml-backward-spaces)
	      (if (not (looking-at "\\s-$"))
		  (1- (current-column))
		(skip-syntax-forward " ")
		(- (current-column) 2))))))
	;; default to DML rules
	(dml-calculate-indentation))))

;;;###autoload
(add-to-list 'auto-mode-alist '("\\.grm\\'" . dml-yacc-mode))
;;;###autoload
(define-derived-mode dml-yacc-mode dml-mode "DML-Yacc"
  "Major Mode for editing ML-Yacc files."
  (set (make-local-variable 'indent-line-function) 'dml-yacc-indent-line)
  (set (make-local-variable 'font-lock-defaults) dml-yacc-font-lock-defaults))

(provide 'dml-mode)

;;; dml-mode.el ends here
