(* dml.sml
 *
 * Copyright (C) 2012  Ian Zerny <ian@zerny.dk>
 *
 * Implementation of Dynamic ML:
 * a dynamically typed language in the ML family.
 * http://www.zerny.dk/dynamic-ml.html
 *)

infixr 3 `			(* internal to the translation *)
infix 4 == != === !==           (* dynamic equalities *)
infix 9 $			(* dynamic application: f$[args...] *)

structure DML = struct

(* internal/raw numbers *)
structure Num = Int

(* the single dynamic type *)
datatype t = VOID
           | BOOL of bool
           | NUM of Num.int
           | STR of string
           | CLO of t list -> t
           | LIST of t list

exception TYPE_CAST_ERROR

(* type predicates *)
fun isVOID d = BOOL (case d of VOID => true | _ => false)
fun isBOOL d = BOOL (case d of BOOL _ => true | _ => false)
fun isNUM d = BOOL (case d of NUM _ => true | _ => false)
fun isSTR d = BOOL (case d of STR _ => true | _ => false)
fun isCLO d = BOOL (case d of CLO _ => true | _ => false)
fun isLIST d = BOOL (case d of LIST _ => true | _ => false)

(* string representation of dynamic values *)
fun REPR v
  = case v
     of VOID => "void"
      | BOOL true  => "true"
      | BOOL false => "false"
      | NUM n => Num.toString n
      | STR s => "'" ^ s ^ "'"
      | CLO f => "<fun>"
      | LIST xs => "[ " ^ String.concatWith ", " (List.map REPR xs) ^ " ]"

(* type casts *)
fun asVOID d = case d of VOID => () | _ => (print ("expected VOID got: "^REPR d); raise TYPE_CAST_ERROR)
fun asBOOL d = case d of BOOL b => b | _ => (print ("expected BOOL got: "^REPR d); raise TYPE_CAST_ERROR)
fun asNUM d = case d of NUM n => n | _ => (print ("expected NUM got: "^REPR d); raise TYPE_CAST_ERROR)
fun asSTR d = case d of STR n => n | _ => (print ("expected STR got: "^REPR d); raise TYPE_CAST_ERROR)
fun asCLO d = case d of CLO f => f | _ => (print ("expected CLO got: "^REPR d); raise TYPE_CAST_ERROR)
fun asLIST d = case d of LIST v => v | _ => (print ("expected LIST got: "^REPR d); raise TYPE_CAST_ERROR)

(* syntax for dynamic application *)
(* read "f$[x, y, z]" as "f(x, y, z)" *)
fun f $ (x:t list) = asCLO f x

(* right infix (native) application *)
(* needed to avoid parenthesis in the translation *)
fun f ` x = f x

(* recursive functions *)
fun fix (f : t) : t
  = let fun f' () = CLO (fn xs => f $ (f' () :: xs))
    in f' ()
    end

(* primitive arithmetic on dynamic values *)
fun op+ (m, n) = NUM (Num.+ (asNUM m, asNUM n))
fun op- (m, n) = NUM (Num.- (asNUM m, asNUM n))
fun op* (m, n) = NUM (Num.* (asNUM m, asNUM n))

(* structural equality *)
local
fun eq (d1, d2) =
    (case (d1, d2)
      of (BOOL v1, BOOL v2) => v1 = v2
       | (NUM v1, NUM v2) => v1 = v2
       | (STR v1, STR v2) => v1 = v2
       | _ => false)
in
fun d1 == d2 = BOOL (eq (d1, d2))
fun d1 != d2 = BOOL (not (eq (d1, d2)))
end

(* printing primitives *)
val PRINT = print o REPR
fun PRINTLN d = (print (REPR d); print "\n")

end


(* A JavaScript-style coercion semantics *)
structure JS = struct
(* Disclaimer:
   This is only suggestive of the JavaScript semantics. 
   There is no NaN number or undefined.
   The string to number conversion doesn't follow JS.
   Etc... *)

local
    open DML
    fun toPrim d
        = (case d
            of CLO _ => STR (REPR d)
             | _ => d)
in

fun asBOOL d
    = (case d
        of VOID   => false
         | BOOL b => b
         | NUM 0  => false
         | NUM _  => true
         | STR "" => false
         | STR _  => true
         | _      => true)

fun asNUM d
    = (case d
        of VOID => 0
         | BOOL true => 1
         | BOOL false => 0
         | NUM n => n
         | STR s => Option.getOpt (Num.fromString s, 0)
         | _ => asNUM (toPrim d))

local 

(* (===) *)
fun eq (d1, d2)
    = (case (d1, d2)
        of (VOID, VOID) => true
         | (BOOL b1, BOOL b2) => b1 = b2
         | (NUM n1, NUM n2) => n1 = n2
         | (STR s1, STR s2) => s1 = s2
         | _ => false)

(* (==) *)
fun equal (d1, d2)
    = (case (d1, d2)
        of (VOID,   VOID)   => eq (d1, d2)
         | (BOOL _, BOOL _) => eq (d1, d2)
         | (NUM _, NUM _) => eq (d1, d2)
         | (STR _, STR _) => eq (d1, d2)
         | (NUM _, STR _) => eq (d1, NUM (asNUM d2))
         | (STR _, NUM _) => eq (NUM (asNUM d1), d2) 
         | (BOOL _, _) => equal (NUM (asNUM d1), d2)
         | (_, BOOL _) => equal (d1, NUM (asNUM d2))
         | _ => false)

fun add_or_concat (d1, d2)
    = (case (toPrim d1, toPrim d2)
        of (STR s, _) => STR (s ^ asSTR d2)
         | (_, STR s) => STR (asSTR d1 ^ s)
         | _ => NUM (Num.+ (asNUM d1, asNUM d2)))

in

fun d1 === d2 = BOOL (eq (d1, d2))
fun d1 !== d2 = BOOL (not (eq (d1, d2)))
fun d1 == d2 = BOOL (equal (d1, d2))
fun d1 != d2 = BOOL (not (equal (d1, d2)))

val op+ = add_or_concat
fun m - n = NUM (Num.- (asNUM m, asNUM n))
fun m * n = NUM (Num.* (asNUM m, asNUM n))

end
end
end;
